module IOC
  class IOCError < ::StandardError; end
  class DuplicateServiceError < IOCError; end
  class RegistrationError < IOCError; end
  class UnknownServiceError < IOCError; end
  class MissingDependenciesError < IOCError; end

  class Container
    def initialize
      @services = {}
      @cache = {}
    end
    
    def register(key, klass=nil, &block)
      raise(DuplicateServiceError, "Service '#{key}' has already been registered") if @services.key?(key)
      raise(RegistrationError, "expected class or block") unless klass || block_given?
      @services[key] = klass ? klass : block
    end

    def resolve(key)
      raise(UnknownServiceError, "Service '#{key}' not registered") unless @services.key?(key)
      
      return @cache[key] if @cache.key?(key)
      
      @cache[key] = 
        if @services[key].is_a?(Proc)
          @services[key].call(self)
        else
          dependencies = get_dependencies(@services[key])
          resolved_dependencies = dependencies.map{|dependency| resolve(dependency)}
          @services[key].new(*resolved_dependencies)
        end
    end

    def resolve_by_type(type)
      key, service = @services.find{|key, service| !service.is_a?(Proc) && service <= type}
      key ? resolve(key) : raise(UnknownServiceError, "Service of type '#{type}' not registered")
    end

    def release
      @cache = {}
    end

    def validate!
      missing = []
      class_services = @services.values.reject{|service| service.is_a?(Proc)}
      class_services.each do |service|
        dependencies = get_dependencies(service)
        dependencies.each do |dependency|
          unless @services.key?(dependency)
            missing << dependency
          end
        end
      end
      missing.uniq!

      unless missing.empty?
        raise(MissingDependenciesError, "missing dependencies: '#{missing.join(', ')}'")
      end
    end

    private

    def get_dependencies(service)
      service.instance_method(:initialize).parameters.map(&:last)
    end
  end
end